#version 150
precision highp float;

//Ethan Alexander Shulman "Cloud Complex" January 15, 2021

in vec3 in_Position;
out float time;
out vec4 uv;

layout(std140) uniform ShaderviewBlock {
	vec4 mouse;
	float screenX,screenY;
	int frame,framerate;
};
#define screen vec2(screenX,screenY)

const vec3 SUN_DIR = normalize(vec3(1));

//#define ANIMATION 1
#define FSAMPLES 100
//#define DEBUG 1
//#define HAS_OPACITY 1

#define IMPORTANCE_SAMPLING 1
#define LIGHT_IMPORTANCE .5
#define LIGHT_POSITION (SUN_DIR*1e6)
#define IMPORTANCE_EMISSION_SCALE 1000.

#define STEPS 100
#define EPS 1e-2
#define REFRACTION_SKIP 5.

#define PI 3.14159265

mat2 r2(float a) {
	float s = sin(a), c = cos(a);
	return mat2(c,-s,s,c);
}

mat2x3 tangentBinormal(vec3 norm) {
	mat2x3 t;
	if (abs(norm.x) > abs(norm.y)) t[0] = normalize(vec3(-norm.z,0,norm.x));
	else t[0] = normalize(vec3(0,norm.z,-norm.y)); 
	t[1] = cross(norm,t[0]);
	return t;
}


vec3 hash(uvec3 x) {
    x = ((x>>8U)^x.yzx)*1103515245U;
    x = ((x>>8U)^x.yzx)*1103515245U;
    x = ((x>>8U)^x.yzx)*1103515245U;
    
    return vec3(x)*(1.0/float(0xffffffffU));
}

vec3 hsvToRGB(vec3 hsv) {
	return mix(clamp(abs(mod(hsv.x*6.0+vec3(0.,4.,2.),6.)-3.0)-1.0,0.,1.),vec3(1),hsv.y)*hsv.z;
}

float twave(float x) {
	return 1.-abs(fract(x)-.5)*2.;
}

float box(vec3 p, vec3 s) {
	p = abs(p)-s;
	return length(max(p,0.))+min(max(p.x,max(p.y,p.z)),0.);
}

float nsampling;

//volumetric opacity dithering
#define START_VOLUME(posSeed) vec3 opacityRand = hash(uvec3(abs(posSeed)*1000.)+uvec3(frame,frame*3,frame*7))
#define VOLUME(dst,skip,alpha,anisotropy) (dst+skip*(.5+opacityRand.y)*step(alpha+nsampling*anisotropy,opacityRand.x))

float geometry(vec3 p) {

	//floor
	float d = p.y+4.;

	//ifs
	vec3 sc = vec3(8,2,8);
	for (int i = 0; i < 4; i++) {
		if (i%2 == 0) d = min(d,box(p,sc*1.3));
		else d = max(d,-box(p,sc));
		p.xz *= r2(float(i));
		p = abs(p)-sc;
		sc *= .5;
	}
	
	//volume dither
	START_VOLUME(p);
	return VOLUME(max(0.,d), .1, .1, 1.);
	
}

//chromatic separation dither
#define START_CHROMA(posSeed) float chromaFrame = float(fract(hash(uvec3(abs(posSeed)*1000.)+uvec3(frame,frame*3,frame*7)).x)*3.)
#define CHROMA(mat,rc,gc,bc) if (chromaFrame < 1.) {mat[0]*=vec3(3,0,0);rc;}\
else if (chromaFrame < 2.) {mat[0]*=vec3(0,3,0);gc;}\
else {mat[0]*=vec3(0,0,3);bc;}

/*mat[0] = color
mat[1].x = forward roughness, mat[1].y = emission, mat[1].z = backward roughness
mat[2].x = opacity, mat[2].y = refraction index
*/
mat3 material(vec3 p) {
	mat3 m = mat3(vec3(.8), vec3(.5,0.,1e-4), vec3(1.,1.,0.));

	return m;
}
vec3 normal(vec3 p) {
	#define NEPS vec3(EPS,0,0)
	nsampling = 1.;
	vec3 n = normalize(vec3(geometry(p+NEPS),geometry(p+NEPS.yxz),geometry(p+NEPS.yzx))-geometry(p));
	nsampling = 0.;
	return n!=n?vec3(0,1,0):n;
}

vec3 background(vec3 rp, vec3 rd) {
	return max(0.,dot(SUN_DIR,rd)-.99)*vec3(50);
}


uniform sampler2D Buffer0;

void vert() {
	//time = float(frame)/float(framerate*FSAMPLES);
	time = 8.63;
	uv = vec4(in_Position.xy*vec2(screenX/screenY,1),(in_Position.xy*.5+.5)*screen);
	gl_Position = vec4(in_Position,1);
}
void frag() {
	float up = pow(twave(time*.1),2.);
	vec3 rp = vec3(0,-3.2+up*8.,-22+up*10.), 
		rd = normalize(vec3(uv.xy+(hash(uvec3(uv.zw,frame)).xy*4.-2.)/screenY,1.)),
		c = vec3(1), l = vec3(0);
		
	mat2 yaw = r2(time*PI*.2);
	rp.xz *= yaw;
	rd.xz *= yaw;
	
	rp += rd*2.;
		
	int i;
	#if HAS_OPACITY
	bool flipRefraction = false;
	#endif
	#if IMPORTANCE_SAMPLING
	bool importanceSample = false,
		importanceHit = false;
	#endif
	nsampling = 0.;
	for (i = 0; i < STEPS; i++) {
		float dst = geometry(rp);
		if (dst <= EPS) {
			mat3 mat = material(rp);
			c *= mat[0];
			#if IMPORTANCE_SAMPLING
			if (importanceSample) {
				l += c*mat[1].y*IMPORTANCE_EMISSION_SCALE;
				importanceHit = true;
				break;
			} else {
				l += c*mat[1].y;
			}
			#else
			l += c*mat[1].y;
			#endif
			
			rp -= rd*(EPS*2.-dst);
			vec3 nrm = normal(rp);
			mat2x3 tangBinorm = tangentBinormal(nrm);
			#if DEBUG
			gl_FragColor = vec4(nrm*.5+.5,1);
			return;
			#endif
			
			vec3 rnd = hash(uvec3(abs(rp)*1000.)+uvec3(frame,frame*3,frame*7));
			#if HAS_OPACITY
			float ct = pow(rnd.x,flipRefraction?mat[1].z:mat[1].x),
			#else
			float ct = pow(rnd.x,mat[1].x),
			#endif
				st = sqrt(1.-ct*ct),
				pt = rnd.y*PI*2.;
			vec3 surf = normalize(nrm*ct+tangBinorm[0]*(st*cos(pt))+tangBinorm[1]*(st*sin(pt)));
			#if HAS_OPACITY
			if (mat[2].x < rnd.z) {
				rp += rd*(EPS*2.-dst);
				rd = refract(rd,surf,flipRefraction?mat[2].y:1./mat[2].y);
				dst = EPS*REFRACTION_SKIP*(1.+rnd.z);
				flipRefraction = !flipRefraction;
			} else {
			#endif
				#if IMPORTANCE_SAMPLING
				mat[1].x = max(1e-3,mat[1].x);
				vec3 refl = mix(reflect(rd,nrm),nrm,pow(mat[1].x,.5)),
					lightDir = normalize(LIGHT_POSITION-rp);
				float light = pow(max(0.,dot(refl,lightDir)),.5/mat[1].x)*50./(1.+mat[1].x*1e4);
				if (rnd.z < LIGHT_IMPORTANCE) {
					//send ray towards light
					c *= light;
					rd = lightDir;
					importanceSample = true;
				} else {
					rd = reflect(rd,surf);
				}
				#else
				rd = reflect(rd,surf);
				#endif
				dst = 0.;
			#if HAS_OPACITY
			}
			#endif
		}
		rp += rd*dst;
	}
	
	//background + final lighting
	#if IMPORTANCE_SAMPLING
	c *= background(rp,rd)*clamp(geometry(rp),0.,1.);
	if (!importanceHit) {
		if (importanceSample) l += c*IMPORTANCE_EMISSION_SCALE;
		else l += c;
	}
	#else
	l += c*background(rp,rd)*clamp(geometry(rp),0.,1.);
	#endif
	
	#if ANIMATION
	gl_FragColor = (frame%FSAMPLES==0?vec4(0):texelFetch(Buffer0,ivec2(uv.zw),0))+vec4(l,1)*1e-2;
	#else
	gl_FragColor = (frame==0?vec4(0):texelFetch(Buffer0,ivec2(uv.zw),0))+vec4(l,1)*1e-2;
	#endif
}
